#ifndef __CDP_H__
#define __CDP_H__

/***************************** interp vars **********************************/
typedef struct {    /* subroutine full address */
//  tptr   code;
  uns32  offset;
     } subracc;

typedef union {
  struct s0
  { uns8   type;
    uns8   val[8];
  } uany;
  struct s1
  { uns8   type;
    double val;
  } udouble;
  struct s2
  { uns8   type;
    sig32 val;
  } usig32;
  struct s3
  { uns8   type;
    monstr val;
  } umonstr;
  struct s4
  { uns8   type;
    tptr   val;
    uns32  valsize;
  } uptr;
} untstr;
#define UNT_SRCDB   1   /* flags for untyped assignment */
#define UNT_DESTDB  2
#define UNT_DESTPAR 4

typedef struct {
  trecnum  position; /* this order is important, position may be used as sig32 */
  tcursnum cursnum;
  tattrib  attr;
  uns16    index;
     } curpos;

typedef union {   /* the basic simple value - 8 bytes */
  sig32   int32;
  byte    money[6];
  monstr  moneys;
  double  real;
  sig16   int16;
  uns8    int8;
  tptr    ptr;
  curpos  wbacc;
  subracc subr;
  untstr * unt;
              } basicval;

#define GR_LEVEL_CNT 6  /* number of aggregating levels, 0=report */

typedef struct aggr_info
{ struct aggr_info * next_info;
  uns16 codesize;
  uns8  type;
  uns8  valsize;
  trecnum count[GR_LEVEL_CNT]; /* number of non-none values */
  char  vls[1];
} aggr_info;

typedef struct
{ aggr_info * aggr_head;
  wbbool  aggr_completed [GR_LEVEL_CNT];
  trecnum aggr_added     [GR_LEVEL_CNT];
  trecnum aggr_count     [GR_LEVEL_CNT];
  uns16   grp_val_offs   [GR_LEVEL_CNT];
    /* total size & partic. offsets of group determining values */
  tptr grp_vals;  /* the group values */
  uns8 glob_level;
} aggr_descr;

typedef struct scopevars    /* space for variables of a scope */
{ struct   scopevars * next_scope;
  unsigned retadr;
  tptr     retcode;
  uns8     wexception;
  unsigned localsize;
  char     vars[1];
} scopevars;

#define I_STACK_SIZE 20
#define MAX_CURS_TABLES   32
#define MAX_KONT_SP  MAX_CURS_TABLES /* must not be less than MAX_CURS_TABLES */

struct objtable;
struct compil_info;

#define MAX_SOURCE_NUMS 50
struct pgm_header_type
{ uns16 version;     // compilation and bytecode version
  uns16 platform;    // platform the EXX has been created on
  uns32 main_entry;  // offset of the program entry point
  unsigned proj_vars_size;   /* -1 if not a base program (for starter) */
  uns32 code_size;   /* incl. header */
  uns32 decl_size;   /* in the file-form */
  uns32 pd_offset;   /* logical offset of the main table */
  int   language;    // the language the program is compiled for
  uns32 compil_timestamp;
  tobjnum sources[MAX_SOURCE_NUMS]; // object numbers of all includes
};

struct global_project_type
{ tobjname   project_name;
  char     * proj_decls;   // owning ptr to global decls 
  objtable * global_decls; // non-owning ptr to the main objtable of global decls
  pgm_header_type * common_proj_code;
  char     * prev_proj_decls;   // owning ptr to invalid global decls 
  objtable * prev_global_decls; // non-owning ptr to the invalid main objtable of global decls
  compil_info * running_program_compilation;
};

/* Values of bp_type */
#define BP_NORMAL      0
#define BP_TEMPORARY   1

typedef struct
{ uns32 bp_offset;
  unsigned bp_line;
  uns8  bp_file;
  uns8  bp_byte;
  uns8  bp_type;
} bp_info;

typedef struct
{ unsigned bp_count;
  unsigned bp_space;  /* total space: existing bps & free cells */
  unsigned current_line;  /* used for single-line stepping & for returning ... */
  uns8  current_file;  /* ... the current line/file on any break */
  BOOL  realized;
  char * code_addr;
  bp_info bps[1];
} breakpoints;

typedef struct
{ uns16 linenum;
  uns8  filenum;
#ifndef WINS
  uns8  padding;
#endif
  uns32 codeoff;
} lnitem;

#define DBGI_ALLOC_STEP  300

typedef struct dbg_info
{ unsigned filecnt;
  unsigned itemcnt;
  unsigned freecells;
#ifndef WINS
  uns8  padding[6];
#endif
  lnitem * lastitem;
  lnitem   items[2];
  void *operator new(unsigned size)
    { return glob_alloc(size); }
  void operator delete(void * ptr)
    { glob_free(ptr); }
  dbg_info(uns32 code_offset, tobjnum src_objnum)
  { freecells=0;  filecnt=1;  itemcnt=1;
    items[0].codeoff=src_objnum;
    items[1].filenum=0;  items[1].linenum=1;  items[1].codeoff=code_offset;
    lastitem=items+1;
  }
  inline tobjnum srcobjs(unsigned filenum)
    { return (tobjnum)items[filenum].codeoff; }
  inline void set_srcobj(unsigned filenum, tobjnum objnum)
    { items[filenum].codeoff=objnum; }
  inline lnitem * lnitems(void)
    { return (lnitem *)(items+filecnt); }
} dbg_info;

#define MAX_CURS_EVID  10
#define MAX_OPEN_FILES  5

/****************************** dynamic array *******************************/
SPECIF_DYNAR(usptr_dynar,untstr*,2);

typedef struct key_trap
{ unsigned keycode;
  unsigned msgnum;
  struct key_trap * next;
  BOOL with_shift;
  BOOL with_ctrl;
  BOOL with_alt;
} key_trap;

//////////////////////////// message queue //////////////////////////////////
// Queue is an array of fixed head (ind. 0) and terminating 0 in msg the field
// after the tail (if not full)

class DllPrezen t_msg_queue
{
 private:
  typedef struct
  { unsigned msg;    /* zero iff terminating item (no message) */
    HWND     hWnd;
  } t_intern_msg;
  enum { INT_MSG_QUEUE_SIZE = 8 };  // queue size
  t_intern_msg queue[INT_MSG_QUEUE_SIZE];

 public:
  t_msg_queue(void) { queue[0].msg=0; }   // not cenessary, done by cdp_init
  void clear(void)  { queue[0].msg=0; }
  BOOL empty(void)  { return !queue[0].msg; }
  BOOL insert       (unsigned msgIn, HWND hWndIn);
  BOOL insert_nodupl(unsigned msgIn, HWND hWndIn);
  BOOL pick_msg     (unsigned * msgOut, HWND * hWndOut);
  void remove_msg   (unsigned msg);
};

///////////////////////////// view synchronization //////////////////////////
#define MAX_REL_ATTRS 4

typedef struct
{ HWND hWndFrom;
  HWND hWndTo;
  UINT ctId;       /* Id of the synchronizing control in the source view */
  tptr dest_expr;  /* NULL for record synchronization, expr for relational */
  int  owner_offset;  /* -1 if synchronized with sel_rec, otherwise synchro with toprec+owner_offset */
  BOOL by_relation;
  tattrib atrsFrom[MAX_REL_ATTRS], // synchronizing attribute numbers in view cursors
          atrsTo  [MAX_REL_ATTRS];
} t_synchro;

SPECIF_DYNAR(synchro_dynar,t_synchro,1);

#define STR_BUF_SIZE 256

typedef struct saved_run_state
{ scopevars    * auto_vars;
  scopevars    * prep_vars;
  scopevars    * glob_vars;
  HGLOBAL      glob_vars_handle;  // not used
  unsigned     pc;
  tptr         runned_code;
  basicval     st[I_STACK_SIZE];
  basicval     * stt;  /* pointer to the stack top (to full cell) */
#ifndef WINS
  LONGJMPBUF   run_jmp;
#endif
  wbbool       mask;
  tcursnum     kont_c[MAX_KONT_SP];
  trecnum      kont_p[MAX_KONT_SP];
  tptr         index_ptr;
  void         * view_ptr;
  basicval     testedval;    /* for constrains testing & history descr. */
  global_project_type run_project;
  uns8         volatile wexception;
  wbbool       prog_running;
  char         hyperbuf[256+2];  /* database access buffer */
  aggr_descr   * aggr;
  uns16        sys_index;    /* must be saved because a view action may open another view */
  uns32        view_position;
  uns16        current_page_number;
  HWND         hWnd;         /* window ID (handle in Windows) */
  wbbool       appended;
  trecnum      append_posit;
  char         strbuf[STR_BUF_SIZE];  /* for strcat & I_MOVESTR */
  char         charstr[2];
  struct saved_run_state * prev_run_state;
} saved_run_state;

typedef struct run_state
{ scopevars    * auto_vars;
  scopevars    * prep_vars;
  scopevars    * glob_vars;
  HGLOBAL      glob_vars_handle;  // not used
  unsigned     pc;
  tptr         runned_code;
  basicval     st[I_STACK_SIZE];
  basicval     * stt;  /* pointer to the stack top (to full cell) */
#ifndef WINS
  LONGJMPBUF   run_jmp;
#endif
  wbbool       mask;
  tcursnum     kont_c[MAX_KONT_SP];
  trecnum      kont_p[MAX_KONT_SP];
  tptr         index_ptr;
  void         * view_ptr;
  basicval     testedval;    /* for constrains testing & history descr. */
  global_project_type run_project;
  uns8         volatile wexception;
  wbbool       prog_running;
  char         hyperbuf[256+2];  /* database access buffer */
  aggr_descr   * aggr;
  uns16        sys_index;
  uns32        view_position;
  uns16        current_page_number;
  HWND         hWnd;         /* window ID (handle in Windows) */
  wbbool       appended;
  trecnum      append_posit;
  char         strbuf[STR_BUF_SIZE];  /* for strcat & I_MOVESTR */
  char         charstr[2];
  saved_run_state * prev_run_state;
 /* not saved vars: */
  void *       files[MAX_OPEN_FILES];
  wbbool       holding;  /* =waiting on server response, used only on client side */
  wbbool       volatile break_request;
  uns8         global_state;
  t_msg_queue  int_msg_queue;  // internal message queue for apps in IPL
  HWND         hClient;
  libinfo      * dllinfo;  /* list of installed DLLs */
  breakpoints  * bpi;
  dbg_info     * dbgi;
  void *       default_tool_bar;
  char   helpfile[80];       /* help file pathname */
  int          report_number;
  uns16        comp_err_line;
  uns16        comp_err_column;
  usptr_dynar  registry;
  key_trap *   key_registry;
  synchro_dynar synchros; // is not destructed id the proper way!!
} run_state;

/********************************** odbc *************************************/
#ifndef WINS
#define WINDOWS
#endif
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>

#define ODBC_TABLE 0x4000
#define IS_ODBC_TABLE(tb) ((tb & (ODBC_TABLE|CURS_USER)) == ODBC_TABLE)


//#ifdef WINS
#define TI_POINTER     1    /* type_info flags */
#define TI_INDEXABLE   2
#define TI_TRACKING    4
#define TI_IS_MONEY    8
#define TI_IS_AUTOINCR 16

typedef struct
{ tptr  strngs; /* ds type name, local type name, prefix, suffix, create pars */
  SWORD sql_type;
  uns32 maxprec;
  sig16 nullable;
  sig16 searchable;
  unsigned flags;
} xtype_info;
CFNC DllKernel int type_WB_2_ODBC(xtype_info * ti, int wbtype);

typedef struct
{ theapsize maxsize;
  char      data[1];
} cached_longobj;
typedef struct
{ theapsize actsize;
  cached_longobj * obj;
} indir_obj;

typedef struct
{ unsigned offset;
  unsigned size;
  uns16    flags;
  uns8     type;
  uns8     multi;
  uns16    odbc_precision;
  uns16    odbc_scale;
  uns16    odbc_sqltype;
  wbbool   changed;  /* changed && identif implies copied */
} atr_info;

struct t_connection;

typedef struct
{ uns8 link_type;  // = LINKTYPE_ODBC
  char dsn[SQL_MAX_DSN_LENGTH+1];
  char tabname  [128+1];
  char owner    [128+1];
  char qualifier[128+1];
} t_odbc_link;

typedef uns16 RPART;
#define BITSPERPART 16

typedef struct DllPrezen ltable
/* cached access structure (for view and for odbc cursor) */
{ cdp_t     cdp;
  ULONG     m_cRef;
  struct t_connection * conn; /* non NULL for ODBC, NULL for WB */
  tcursnum  cursnum;
  HSTMT     hStmt;            /* for ODBC only */
  uns32     supercurs;        // upper cursnum (WB) or hStmt (ODBC)
  BOOL      close_cursor;     // closing cursor requred when deleting (WB only)
  tptr      select_st;        /* for ODBC an cursor reset only */
 // cache:
  trecnum   cache_top;        /* number of the record in cache(top), NORECNUM iff none */
  unsigned  cache_reccnt;
  LONG      rec_cache_size;    /* actual cache buffer size, >= rec_cache_real */
  unsigned  rec_cache_real;    /* real record size */
  HGLOBAL   cache_hnd;         /* cache buffer handle */
  tptr      cache_ptr;     /* cache pointer */
  tptr      odbc_buf;      /* aux ODBC buffer for 1 record, used by BindColumn */
  trecnum   edt_rec;   /* number of record being edited, NORECNUM iff synchronized */
  unsigned  rr_datasize;
  unsigned  attrcnt;
  atr_info * attrs;
  BOOL      recprivs;
  unsigned  rec_priv_off;  // offset of record privils in the cache, valid iff recprivs
 /* curspor position, not used by views: */

 // remap:
  RPART *   remap_a;       /* remap bitmap or NULL iff no remap */
  HGLOBAL   remap_handle;  /* handle of remap_a or NULL iff no remap */
  uns32     remap_size;    /* size of remap_a in bytes */
  trecnum   rm_ext_recnum; /* number of ext. records covered in remap_a */
  trecnum   rm_int_recnum; /* number of remapped internal records */
  wbbool    rm_completed;  /* wbtrue iff remap_a contains all records */
  trecnum   rm_curr_ext;   /* current record: external number */
  trecnum   rm_curr_int;   /* current record: internal number */
  trecnum   fi_recnum;     /* replaces "recnum" */

  ltable(cdp_t cdpIn, struct t_connection * connIn);
  ~ltable(void);

  ULONG AddRef(void) { return ++m_cRef; }
  ULONG Release(void){ if (--m_cRef) return m_cRef; delete this; return 0; }
} ltable;

typedef struct  /* registered odbc table or cursor */
{ tcursnum  odbc_num;
  BOOL      is_table;
  BOOL      mapped;
  union
  { tptr          source;     /* select statement for cursors (!is_table) */
    t_odbc_link * odbc_link;  /* iff is_table */
  }         s;
  struct t_connection * conn;
  ltable *  ltab;      /* used for accessing data */
} odbc_tabcurs;

SPECIF_DYNAR(odbc_tc_dynar,odbc_tabcurs,1);

#define CONN_FL_TEMPORARY               1
#define CONN_FL_NAMED_CONSTRS           2
#define CONN_FL_NON_NULL                4
#define CONN_FL_DISABLE_COLUMN_PRIVILS  8

typedef struct t_connection
{ t_connection * next_conn;
  BOOL  selected;
  tobjnum objnum;  // WB object number, GUI uses it
  char  dsn[SQL_MAX_DSN_LENGTH+1];
  tptr  conn_string;
  HDBC  hDbc;
  HSTMT hStmt;
  uns32 owner_usage, qualifier_usage;
  uns16 qualifier_loc;
  char  qualifier_separator[5];
  char  identifier_quote_char;
  BOOL  SetPos_support;   /* UPDATE, ADD and DELETE supported in SetPos */
  BOOL  can_lock;
  BOOL  absolute_fetch;
  BOOL  can_transact;
  uns32 scroll_options, scroll_concur;
  uns32 flags;
  uns16 sql_conform;
  xtype_info * ti;
  odbc_tc_dynar ltabs;
  inline t_connection:: t_connection(void)
    { conn_string=NULL;  ti=NULL;  selected=FALSE;  objnum=(tobjnum)-1; }
  inline t_connection::~t_connection(void)
    { corefree(conn_string); corefree(ti); }
} t_connection;

typedef struct  // ODBC connection information stored in a connection object
{ char dsname[SQL_MAX_DSN_LENGTH+1];
  char conn_string[1024];
} t_conndef;

//#endif
struct t_folder_header
{ uns32 version;
};
struct t_folder_item
{ tobjname name;
  tcateg   categ;
  uns32    modif_timestamp;
};

typedef struct
{ tobjname name;
  tobjnum objnum;
  uns16   flags;
  t_folder_index folder;
  uns32          modif_timestamp;
} t_comp;

SPECIF_DYNAR(comp_dynar, t_comp, 6);

struct t_relation : public t_comp
{ uns8     refint, index1, index2, nic;
  tobjname object_folder; // duplicates info in "folder" but is necessary to be consistent (?)
  tobjname tab1name;
  tobjname tab2name;
  tobjnum  tab1num, tab2num;
  tattrib  atr1[MAX_REL_ATTRS];
  tattrib  atr2[MAX_REL_ATTRS];
};

SPECIF_DYNAR(relation_dynar, t_relation, 3);

typedef t_comp t_folder;
//struct t_folder : public t_comp
//{ tcateg fcateg;
//  HANDLE hTreeItem;
//};

SPECIF_DYNAR(folder_dynar, t_folder, 3);

struct t_odbc
{ 
#ifdef WINS
  HENV hEnv;
  t_connection * conns;
#endif
  BOOL mapping_on;
  comp_dynar      apl_tables;
  comp_dynar      apl_views;
  comp_dynar      apl_menus;
  comp_dynar      apl_queries;
  comp_dynar      apl_programs;
  comp_dynar      apl_pictures;
  comp_dynar      apl_drawings;
  comp_dynar      apl_replrels;
  comp_dynar      apl_wwws;
  relation_dynar  apl_relations;
  comp_dynar      apl_procs;
  comp_dynar      apl_triggers;
  comp_dynar      apl_seqs;
  folder_dynar    apl_folders;
};

///////////////////// embedded vars, prepared stmts ///////////////////////

struct t_clivar
{ tname name;
  object * obj;
  t_parmode mode;
  int type;
  void * buf;
  int buflen;
};

SPECIF_DYNAR(t_clivar_dynar, t_clivar, 2);

struct t_prep_stmt
{ BOOL is_prepared;
  t_clivar_dynar clivars;
};

SPECIF_DYNAR(t_prep_dynar, t_prep_stmt, 1);

/****************************** cdp ******************************************/
struct view_dyn;
struct IStorage;
class cAddress;

struct sBuffDescr // client request buffer element structure
{ unsigned     Len;
  const char * pBuff;
};

typedef struct cd_t
{ byte   in_use;  /* is PT_EMPTY, PT_DIRECT, PT_REMOTE, PT_SLAVE, PT_SERVER */
  uns8   server_index; // user in dual-access functions, may be removed in 32-bits (if 1 client cannot link to 2 servers simult.)
  wbbool mask_token;  // true if write allowed without the token
  wbbool non_mdi_client;
 // description of the connection between client and server:
#ifdef UNIX
  LocEvent hSlaveSem;   
#else
	HANDLE   hSlaveSem;   // not used!
#endif
	DWORD		 LinkIdNumber;  /* used only by the direct client on Break */
 // selected application:
  tobjname    sel_appl_name;  // NULLSTRING on no application selected
  WBUUID      sel_appl_uuid;
  WBUUID      front_end_uuid; // same as sel_appl_uuid if not redirected
  WBUUID      back_end_uuid;  // same as sel_appl_uuid if not redirected
  tobjnum     admin_role;     // NOOBJECT iff role does not exist
  tobjnum     senior_role;
  tobjnum     junior_role;
  wbbool      locked_appl;    // enciphered and editing disabled

  uns8        applnum;  /* number identifying the connection (in d_cached rights) */
  run_state   RV;
  MsgInfo     msg;          // incomming message (server answer) - released after the write_ans_proc is called
  uns8        reqBuffType;  // type of buffer in the "msg"
  t_thread    threadID;     // thread or task ID
  cAddress *  pRemAddr;     // connection address, assigned in interf_init_ex, deleted in Unlink
  BYTE *      pImpExpBuff;  // not used!
  WORD        wImpExpDataSize;  // not used!
  uns32       auxinfo; // container, client linking result, data import flag
  view_dyn *  all_views;    // views list on client, dest server on slave
  tobjname    errmsg;

 // client only variables:
  tobjnum     applobj;  /* application object number (on the client side) */
  void *      OleMsgFilter;
  t_odbc      odbc;
  long        server_version;
 // client comm. vars:
  tobjname conn_server_name; // stored by interf_init, used by cd_unlink_kernel
  sBuffDescr  req_array[MAX_PACKAGED_REQS+1]; // client request in parts
  void      * ans_array[MAX_PACKAGED_REQS]; // pointers for the results
  uns16       ans_type [MAX_PACKAGED_REQS];
  wbbool      in_package, in_concurrent;
  uns8        req_ind, ans_ind;
  uns16    lastkey;          /* the first part of each request */
  char     default_rq_frame[PACKET_PREFIX_SIZE+DEFAULT_RQ_FRAME_SIZE+1];
  wbbool   default_frame_busy;
  wbbool   last_error_is_from_client;
 // direct client synchonization handles 
#ifdef WINS
  HANDLE hDataEvent;
  HANDLE hDoneEvent;
  HANDLE hReqMap;
  HANDLE hAnsMap;
#else
  t_dircomm dircomm;
#endif
 // saved private key:
  t_priv_key_file * priv_key_file;
  t_prep_dynar prep_stmts;
  uns32    login_key;
  wbbool global_debug_mode;
 // multilingual support:
  ltable * lang_cache;
  trecnum cache_recs;
  int selected_lang;
#ifdef WINS
  IStorage * IStorageCache;
#endif
  WBUUID   server_uuid;
  uns32    obtained_login_key;  // login key obtained by this client and registered
} cd_t;

#endif  /* ! __CDP_H__ */

